/**
 * PETALS - PETALS Services Platform. Copyright (c) 2007 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.component.framework;

import java.lang.reflect.Field;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;

import javax.management.Attribute;
import javax.management.AttributeList;
import javax.management.AttributeNotFoundException;
import javax.management.DynamicMBean;
import javax.management.IntrospectionException;
import javax.management.InvalidAttributeValueException;
import javax.management.MBeanAttributeInfo;
import javax.management.MBeanException;
import javax.management.MBeanInfo;
import javax.management.MBeanOperationInfo;
import javax.management.ReflectionException;
import javax.management.RuntimeOperationsException;

import org.ow2.petals.component.framework.jbidescriptor.generated.Component;

/**
 * The dynamic MBean used to expose component attributes and operations via JMX.
 * 
 * @author chamerling - EBM WebSourcing
 * @author rnaudin - EBM WebSourcing
 */
public class ExtensionMBean implements DynamicMBean {

    private static class AttributeSetterGetter {

        public String attributeName;

        public Method getter;

        public Method setter;

        public AttributeSetterGetter(final String attributeName, final Method getter,
                final Method setter) {
            this.attributeName = attributeName;
            this.getter = getter;
            this.setter = setter;
        }
    }

    private final List<AttributeSetterGetter> attributeSetterGetterList = new ArrayList<AttributeSetterGetter>();

    private final static Hashtable<String, Class<?>> primitiveClasses = new Hashtable<String, Class<?>>(
            8);

    private final DefaultBootstrap bootstrap;
    {
        primitiveClasses.put(Boolean.TYPE.toString(), Boolean.TYPE);
        primitiveClasses.put(Character.TYPE.toString(), Character.TYPE);
        primitiveClasses.put(Byte.TYPE.toString(), Byte.TYPE);
        primitiveClasses.put(Short.TYPE.toString(), Short.TYPE);
        primitiveClasses.put(Integer.TYPE.toString(), Integer.TYPE);
        primitiveClasses.put(Long.TYPE.toString(), Long.TYPE);
        primitiveClasses.put(Float.TYPE.toString(), Float.TYPE);
        primitiveClasses.put(Double.TYPE.toString(), Double.TYPE);
    }

    /**
     * Creates a new instance of {@link ExtensionMBean}
     * 
     * @param bootstrap
     */
    public ExtensionMBean(final DefaultBootstrap bootstrap) {
        this.bootstrap = bootstrap;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.management.DynamicMBean#getAttribute(java.lang.String)
     */
    public synchronized Object getAttribute(final String attribute)
    throws AttributeNotFoundException, MBeanException, ReflectionException {
        Object attributeObject = null;

        try {
            final Field field = Component.class.getDeclaredField(attribute);
            field.setAccessible(true);
            final Object settableObject = field.get(this.bootstrap.getJbiComponentConfiguration()
                    .getComponent());
            if (settableObject != null) {
                String methodName = "getValue";
                if (field.getType().getSimpleName().endsWith("boolean")) {
                    methodName = "isValue";
                }
                final Method method = settableObject.getClass().getMethod(methodName);
                attributeObject = method.invoke(settableObject);
            }

        } catch (final SecurityException e) {
            throw new AttributeNotFoundException("Attribute '" + attribute + "' is not reachable");
        } catch (final NoSuchFieldException e) {
            // try to get a custom attribute
            boolean isFound = false;
            try {
                for (final AttributeSetterGetter attributeSetterGetter : this.attributeSetterGetterList) {
                    if (attributeSetterGetter.attributeName.equals(attribute)) {
                        attributeObject = attributeSetterGetter.getter.invoke(this.bootstrap);
                        isFound = true;
                        break;
                    }
                }
            } catch (final IllegalArgumentException e1) {
                throw new ReflectionException(e, e.getMessage());
            } catch (final IllegalAccessException e1) {
                throw new ReflectionException(e, e.getMessage());
            } catch (final InvocationTargetException e1) {
                throw new ReflectionException(e, e.getMessage());
            }

            if (isFound == false) {
                throw new AttributeNotFoundException("Attribute '" + attribute + "' does not exist");
            }
        } catch (final IllegalArgumentException e) {
            throw new ReflectionException(e, e.getMessage());
        } catch (final IllegalAccessException e) {
            throw new ReflectionException(e, e.getMessage());
        } catch (final NoSuchMethodException e) {
            throw new ReflectionException(e, e.getMessage());
        } catch (final InvocationTargetException e) {
            throw new ReflectionException(e, e.getMessage());
        }

        return attributeObject;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.management.DynamicMBean#getAttributes(java.lang.String[])
     */
    public synchronized AttributeList getAttributes(final String[] attributes) {
        final AttributeList list = new AttributeList();

        for (final String attribute : attributes) {
            try {
                list.add(new Attribute(attribute, this.getAttribute(attribute)));
            } catch (final AttributeNotFoundException e) {
                this.bootstrap.getLogger().warning(
                        "Failed to get attribute '" + attribute + "' in the extension MBean : "
                        + e.getClass().getSimpleName() + " : " + e.getMessage());
            } catch (final MBeanException e) {
                this.bootstrap.getLogger().warning(
                        "Failed to get attribute '" + attribute + "' in the extension MBean : "
                        + e.getClass().getSimpleName() + " : " + e.getMessage());
            } catch (final ReflectionException e) {
                this.bootstrap.getLogger().warning(
                        "Failed to get attribute '" + attribute + "' in the extension MBean : "
                        + e.getClass().getSimpleName() + " : " + e.getMessage());
            }
        }

        return list;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * javax.management.DynamicMBean#setAttribute(javax.management.Attribute)
     */
    public synchronized void setAttribute(final Attribute attribute)
    throws AttributeNotFoundException, InvalidAttributeValueException, MBeanException,
    ReflectionException {

        try {
            final Field field = Component.class.getDeclaredField(attribute.getName());
            field.setAccessible(true);
            final Object settableObject = field.get(this.bootstrap.getJbiComponentConfiguration()
                    .getComponent());
            final Field primitiveField = field.getType().getDeclaredField("value");
            final Method method = settableObject.getClass().getMethod("setValue",
                    new Class[] { primitiveField.getType() });
            method.invoke(settableObject, new Object[] { attribute.getValue() });
        } catch (final SecurityException e) {
            throw new AttributeNotFoundException("Attribute '" + attribute.getName()
                    + "' is not reachable");
        } catch (final NoSuchFieldException e) {
            // try to set a custom attribute
            boolean isFound = false;
            try {
                for (final AttributeSetterGetter attributeSetterGetter : this.attributeSetterGetterList) {
                    if (attributeSetterGetter.attributeName.equals(attribute.getName())) {
                        attributeSetterGetter.setter.invoke(this.bootstrap, attribute.getValue());
                        isFound = true;
                        break;
                    }
                }
            } catch (final IllegalArgumentException e1) {
                throw new ReflectionException(e, e.getMessage());
            } catch (final IllegalAccessException e1) {
                throw new ReflectionException(e, e.getMessage());
            } catch (final InvocationTargetException e1) {
                throw new ReflectionException(e, e.getMessage());
            }

            if (isFound == false) {
                throw new AttributeNotFoundException("Attribute '" + attribute + "' does not exist");
            }
        } catch (final IllegalArgumentException e) {
            throw new ReflectionException(e);
        } catch (final IllegalAccessException e) {
            throw new ReflectionException(e);
        } catch (final NoSuchMethodException e) {
            throw new ReflectionException(e);
        } catch (final InvocationTargetException e) {
            throw new ReflectionException(e);
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * javax.management.DynamicMBean#setAttributes(javax.management.AttributeList
     * )
     */
    public synchronized AttributeList setAttributes(final AttributeList list) {
        final AttributeList attributeList = new AttributeList();

        for (final Object attribute : list) {
            try {
                try {
                    final Field field = Component.class.getDeclaredField(((Attribute) attribute)
                            .getName());
                    field.setAccessible(true);
                    final Object settableObject = field.get(this.bootstrap
                            .getJbiComponentConfiguration().getComponent());
                    final Field primitiveField = field.getType().getDeclaredField("value");
                    final Method method = settableObject.getClass().getMethod("setValue",
                            new Class[] { primitiveField.getType() });
                    method.invoke(settableObject,
                            new Object[] { ((Attribute) attribute).getValue() });
                } catch (final NoSuchFieldException e) {
                    // try to set a custom attribute
                    boolean isFound = false;
                    for (final AttributeSetterGetter attributeSetterGetter : this.attributeSetterGetterList) {
                        if (attributeSetterGetter.attributeName.equals(((Attribute) attribute)
                                .getName())) {
                            attributeSetterGetter.setter.invoke(this.bootstrap,
                                    ((Attribute) attribute).getValue());
                            isFound = true;
                            break;
                        }
                    }

                    if (isFound == false) {
                        throw new AttributeNotFoundException("Attribute '" + attribute
                                + "' does not exist");
                    }
                }
            } catch (final Throwable t) {
                this.bootstrap.getLogger().warning(
                        "Failed to set attribute '" + ((Attribute) attribute).getName()
                        + "' in the extension MBean : " + t.getClass().getSimpleName()
                        + " : " + t.getMessage());
            }
        }

        return attributeList;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.management.DynamicMBean#invoke(java.lang.String,
     * java.lang.Object[], java.lang.String[])
     */
    public Object invoke(final String actionName, final Object[] params, final String[] signature)
    throws MBeanException, ReflectionException {
        Object result = null;
        Class<?>[] parameters = null;

        final Class<?> objClass = this.bootstrap.getClass();
        final ClassLoader loader = objClass.getClassLoader();
        if (signature != null) {
            parameters = this.findSignatureClasses(signature, loader);
        }
        // Query the metadata service to get the right method
        //
        Method method = null;
        try {
            method = objClass.getMethod(actionName, parameters);
        } catch (final Throwable t) {
            // Nothing to do
        }

        if (method == null) {
            throw new ReflectionException(new NoSuchMethodException(actionName),
                    "The operation with name " + actionName + " could not be found");
        }

        try {
            // invoke method
            result = method.invoke(this.bootstrap, params);
        } catch (final IllegalAccessException e) {
            throw new ReflectionException(e, "IllegalAccessException"
                    + " occured trying to invoke operation " + actionName);
        } catch (final RuntimeException e) {
            throw new RuntimeOperationsException(e, "RuntimeException"
                    + " occured trying to invoke operation " + actionName);
        } catch (final InvocationTargetException e) {
            throw new MBeanException(e, "Exception thrown in operation " + actionName);
        }
        return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @see javax.management.DynamicMBean#getMBeanInfo()
     */
    public synchronized MBeanInfo getMBeanInfo() {
        // get the attributes
        final MBeanAttributeInfo[] attrs = this.getAttributesInfos();
        // get the operations
        final MBeanOperationInfo[] opps = this.getOperationsInfos();

        // create MBeanInfo with no constructors and no notifications
        return new MBeanInfo(this.getClass().getName(), "Property Manager MBean", attrs, null,
                opps, null);
    }

    /**
     * Dynamically get attributes for the MBean. The attributes are the ones
     * from the properties file.
     * 
     * @return
     */
    protected MBeanAttributeInfo[] getAttributesInfos() {
        MBeanAttributeInfo[] attrs = null;
        final Field[] fields = Component.class.getDeclaredFields();
        final List<Field> settableFields = new ArrayList<Field>();
        final Component componentConfiguration = this.bootstrap.getJbiComponentConfiguration()
        .getComponent();

        try {
            // get the settable fields
            for (final Field field : fields) {
                if (field.getType().getSimpleName().startsWith("Runtime")
                        || field.getType().getSimpleName().startsWith("Settable")) {
                    field.setAccessible(true);
                    if (field.get(componentConfiguration) != null) {
                        settableFields.add(field);
                    }
                }
            }

            // get the custom attributes list
            final List<String> customAttrs = this.bootstrap.getAttributeList();

            // build the MBeanAttributeInfo
            int count = 0;
            attrs = new MBeanAttributeInfo[settableFields.size() + customAttrs.size()];

            final Method[] methodList = this.bootstrap.getClass().getMethods();
            for (final String attr : customAttrs) {
                Method setter = null;
                Method getter = null;
                for (final Method method : methodList) {
                    if (method.getName().equalsIgnoreCase("set" + attr)) {
                        setter = method;
                    } else if (method.getName().equalsIgnoreCase("get" + attr)
                            || method.getName().equalsIgnoreCase("is" + attr)) {
                        getter = method;
                    }
                }
                attrs[count] = new MBeanAttributeInfo(attr, "Property " + attr, getter, setter);
                this.attributeSetterGetterList.add(new AttributeSetterGetter(attr, getter, setter));
                count++;
            }

            for (final Field field : settableFields) {
                final Object settableObject = field.get(componentConfiguration);
                Method getter;
                if (field.getType().getSimpleName().endsWith("boolean")) {
                    getter = settableObject.getClass().getMethod("isValue");
                } else {
                    getter = settableObject.getClass().getMethod("getValue");
                }
                final Field primitiveField = field.getType().getDeclaredField("value");
                final Method setter = settableObject.getClass().getMethod("setValue",
                        new Class[] { primitiveField.getType() });

                attrs[count] = new MBeanAttributeInfo(field.getName(), "Property "
                        + field.getName(), getter, setter);
                count++;

            }

        } catch (final IllegalArgumentException e) {
            this.bootstrap.getLogger().warning(
                    "Failed to initialize fields of the Bootstrap extension MBean : "
                    + e.getClass().getSimpleName() + " : " + e.getMessage());
        } catch (final IllegalAccessException e) {
            this.bootstrap.getLogger().warning(
                    "Failed to initialize fields of the Bootstrap extension MBean : "
                    + e.getClass().getSimpleName() + " : " + e.getMessage());
        } catch (final SecurityException e) {
            this.bootstrap.getLogger().warning(
                    "Failed to initialize fields of the Bootstrap extension MBean : "
                    + e.getClass().getSimpleName() + " : " + e.getMessage());
        } catch (final NoSuchMethodException e) {
            this.bootstrap.getLogger().warning(
                    "Failed to initialize fields of the Bootstrap extension MBean : "
                    + e.getClass().getSimpleName() + " : " + e.getMessage());
        } catch (final IntrospectionException e) {
            this.bootstrap.getLogger().warning(
                    "Failed to initialize fields of the Bootstrap extension MBean : "
                    + e.getClass().getSimpleName() + " : " + e.getMessage());
        } catch (final NoSuchFieldException e) {
            this.bootstrap.getLogger().warning(
                    "Failed to initialize fields of the Bootstrap extension MBean : "
                    + e.getClass().getSimpleName() + " : " + e.getMessage());
        }

        return attrs;
    }

    /**
     * Dynamically get the operations for the MBean. The operations are defined
     * in the properties manager with a special prefix.
     * 
     * @return
     */
    protected MBeanOperationInfo[] getOperationsInfos() {
        final List<String> methods = this.bootstrap.getMethodList();
        final List<MBeanOperationInfo> operations = new ArrayList<MBeanOperationInfo>();
        final Method[] methodList = this.bootstrap.getClass().getMethods();

        for (final Method method : methodList) {
            final String name = method.getName();

            // add method if it has been defined by the user
            if (methods.contains(name)) {
                final MBeanOperationInfo oper = new MBeanOperationInfo("Operation description",
                        method);
                operations.add(oper);
            }
        }

        final MBeanOperationInfo[] result = new MBeanOperationInfo[operations.size()];
        operations.toArray(result);

        return result;
    }

    /**
     * Find the classes of the signature.
     * 
     * @param signature
     * @param loader
     * @return
     * @throws ReflectionException
     */
    private Class<?>[] findSignatureClasses(final String signature[], final ClassLoader loader)
    throws ReflectionException {
        if (signature == null) {
            return null;
        }

        final int length = signature.length;
        final Class<?> tab[] = new Class[length];

        if (length == 0) {
            return tab;
        }

        try {
            for (int i = 0; i < length; i++) {
                // try to get the primitive class
                final Class<?> primCla = primitiveClasses.get(signature[i]);
                if (primCla != null) {
                    tab[i] = primCla;
                } else {
                    // try to get the class if not primitive one
                    tab[i] = Class.forName(signature[i], false, loader);
                }
            }
        } catch (final ClassNotFoundException e) {
            throw new ReflectionException(e, "The parameter class could not be found");
        } catch (final RuntimeException e) {
            throw e;
        }
        return tab;
    }

}
