/**
 * PETALS - PETALS Services Platform. Copyright (c) 2005 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * ---------------------------------------------------------------------------
 * ---------------------------------------------------------------------------
 */
package org.ow2.petals.component.framework.notification;

import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;
import java.util.logging.Logger;

import javax.jbi.messaging.DeliveryChannel;
import javax.jbi.messaging.MessagingException;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactoryConfigurationError;
import javax.xml.xpath.XPathExpressionException;

import org.apache.commons.lang.NotImplementedException;
import org.ow2.easywsdl.wsdl.api.WSDLException;
import org.ow2.petals.component.framework.api.Constants;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.notification.NotificationBrokerController;
import org.ow2.petals.component.framework.api.notification.NotificationProducer;
import org.ow2.petals.component.framework.api.notification.notify.NotifyController;
import org.ow2.petals.component.framework.notification.notify.ComponentNotifyController;

import com.ebmwebsourcing.wsstar.addressing.definition.WSAddressingFactory;
import com.ebmwebsourcing.wsstar.addressing.definition.api.EndpointReferenceType;
import com.ebmwebsourcing.wsstar.addressing.definition.api.ReferenceParametersType;
import com.ebmwebsourcing.wsstar.addressing.definition.api.WSAddressingException;
import com.ebmwebsourcing.wsstar.notification.definition.WSNotificationFactory;
import com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.GetCurrentMessage;
import com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.GetCurrentMessageResponse;
import com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.Renew;
import com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.RenewResponse;
import com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.Subscribe;
import com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.SubscribeResponse;
import com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.Unsubscribe;
import com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.UnsubscribeResponse;
import com.ebmwebsourcing.wsstar.notification.definition.utils.WSNotificationException;
import com.ebmwebsourcing.wsstar.notification.extension.WSNotificationExtensionFactory;
import com.ebmwebsourcing.wsstar.notification.extension.api.ResourcesUuidType;
import com.ebmwebsourcing.wsstar.notification.extension.api.SOAParameterType;
import com.ebmwebsourcing.wsstar.notification.extension.utils.WSNotificationExtensionException;
import com.ebmwebsourcing.wsstar.notification.extension.utils.WsnSpecificTypeHelper;
import com.ebmwebsourcing.wsstar.notification.service.fault.WSNotificationFault;

/**
 * @author Frederic Gardes
 * 
 */
public class ComponentNotificationProducer extends AbstractComponentNotificationActor implements
NotificationProducer {

    private Map<EndpointReferenceType, Subscribe> subscriptionMap = null;

    /**
     * The controller of Component Notify
     */
    private NotifyController componentNotifyController = null;

    /**
     * The constructor
     * 
     * @param deliveryChannel
     *            The delivery channel
     * @param wsnWriter
     * @param notificationBrokerController
     * @throws WSDLException
     * @throws WSNotificationException
     * @throws WSAddressingException
     */
    public ComponentNotificationProducer(final DeliveryChannel deliveryChannel,

            final Logger logger, final NotificationBrokerController notificationBrokerController)
    throws WSDLException, WSNotificationException, WSAddressingException {
        super(Constants.Notification.PRODUCER_WSDL, logger);
        this.componentNotifyController = new ComponentNotifyController(
                notificationBrokerController, logger);
        this.subscriptionMap = new ConcurrentHashMap<EndpointReferenceType, Subscribe>();
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.notification.NotificationProducer
     * #getCurrentMessage
     * (com.ebmwebsourcing.wsstar.notification.definition.basenotification
     * .api.GetCurrentMessage)
     */
    public GetCurrentMessageResponse getCurrentMessage(final GetCurrentMessage getCurrentMessage) {
        throw new NotImplementedException();
    }


    public NotifyController getNotifyController() {
        return this.componentNotifyController;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.notification.NotificationProducer
     * #getNotificationProducerSOAParameter()
     */
    public SOAParameterType getNotificationProducerSOAParameter() throws WSDLException,
    WSNotificationException, WSAddressingException, WSNotificationExtensionException {
        return this.getSOAParameter(Constants.Notification.NOTIFICATION_PRODUCER_INTERFACE);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.notification.NotificationProducer
     * #getSubscriptionManagerSOAParameter()
     */
    public SOAParameterType getSubscriptionManagerSOAParameter() throws WSDLException,
    WSNotificationException, WSNotificationExtensionException, WSAddressingException {
        return this.getSOAParameter(Constants.Notification.SUBSCRIPTION_INTERFACE);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.notification.NotificationProducer
     * #getSubscriptionMap()
     */
    public Map<EndpointReferenceType, Subscribe> getSubscriptionMap() {
        return this.subscriptionMap;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.notification.NotificationProducer
     * #renew
     * (com.ebmwebsourcing.wsstar.notification.definition.basenotification.
     * api.Renew)
     */
    public RenewResponse renew(final Renew renew) {
        throw new NotImplementedException();
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.notification.NotificationProducer
     * #subscribe
     * (com.ebmwebsourcing.wsstar.notification.definition.basenotification
     * .api.Subscribe)
     */
    public SubscribeResponse subscribe(final Subscribe subscribe) throws PEtALSCDKException,
    WSNotificationException, WSNotificationExtensionException, WSAddressingException,
    WSDLException, MessagingException, WSNotificationFault, XPathExpressionException,
    TransformerFactoryConfigurationError, TransformerException,
    ParserConfigurationException {
        // subscribe response
        final SubscribeResponse subscribeResponse = WSNotificationFactory.getInstance()
        .createSubscribeResponse();
        // Subscription Reference
        final String uuid = /* "uuid:" + */UUID.randomUUID().toString();
        final EndpointReferenceType subscriptionReference = WSAddressingFactory.getInstance()
        .newEndpointReferenceType();
        // address
        subscriptionReference.setAddress(Constants.Notification.CDK_ADDRESS);
        // referenceParameters with a SOA parameter and a resource uuid
        final SOAParameterType soaParameter = this.getSubscriptionManagerSOAParameter();
        final ReferenceParametersType referenceParametersType = subscriptionReference
        .newReferenceParameters();
        WsnSpecificTypeHelper.setSOAParameter(soaParameter, referenceParametersType);
        final ResourcesUuidType resourcesUuidType = WSNotificationExtensionFactory.getInstance()
        .createResourcesUuidType();
        resourcesUuidType.addUuid(uuid);
        WsnSpecificTypeHelper.setResourcesUuidType(resourcesUuidType, referenceParametersType);
        subscriptionReference.setReferenceParameters(referenceParametersType);
        subscribeResponse.setSubscriptionReference(subscriptionReference);
        // Current Time
        // Termination Time
        // add the Subscribe to the subscription map
        this.subscriptionMap.put(subscriptionReference, subscribe);
        // add the Subscribe to the component notify controller
        this.componentNotifyController.add(subscriptionReference, subscribe);
        return subscribeResponse;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.api.notification.NotificationProducer
     * #unsubscribe
     * (com.ebmwebsourcing.wsstar.notification.definition.basenotification
     * .api.Unsubscribe)
     */
    public UnsubscribeResponse unsubscribe(final Unsubscribe unsubscribe)
    throws PEtALSCDKException, WSNotificationException, WSNotificationExtensionException,
    WSAddressingException, MessagingException, WSNotificationFault {
        boolean found = false;
        String uuid = null;
        ResourcesUuidType resourcesUuid = null;

        // remove the subscribe on the Map
        if (unsubscribe != null) {
            resourcesUuid = WsnSpecificTypeHelper.getResourcesUuidType(unsubscribe);
            if (resourcesUuid != null) {
                final List<String> resourcesUuids = resourcesUuid.getUuids();
                if (resourcesUuids != null) {
                    final Iterator<EndpointReferenceType> iterator = this.subscriptionMap.keySet()
                    .iterator();
                    while (iterator.hasNext() && !resourcesUuids.isEmpty()) {
                        final EndpointReferenceType endpointReferenceType = iterator.next();
                        if (endpointReferenceType != null
                                && endpointReferenceType.getReferenceParameters() != null) {
                            resourcesUuid = WsnSpecificTypeHelper
                            .getResourcesUuidType(endpointReferenceType
                                    .getReferenceParameters());
                            if (resourcesUuid != null && resourcesUuid.getUuids() != null) {
                                if (resourcesUuid.getUuids().size() != 1) {
                                    throw new PEtALSCDKException(
                                    "Unable to unsubscribe a Subscription Reference without one and only one resource uuid");
                                }
                                final String storedUuid = resourcesUuid.getUuids().get(0);
                                final Iterator<String> resourcesUuidsIterator = resourcesUuids
                                .iterator();
                                found = false;
                                while (resourcesUuidsIterator.hasNext() && !found) {
                                    uuid = resourcesUuidsIterator.next();
                                    if (storedUuid.equals(uuid)) {
                                        found = true;
                                        resourcesUuids.remove(storedUuid);
                                        this.componentNotifyController
                                        .remove(endpointReferenceType);
                                        this.subscriptionMap.remove(endpointReferenceType);
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
        // subscribe response
        return WSNotificationFactory.getInstance().createUnsubscribeResponse();
    }
}
