/**
 * PETALS - PETALS Services Platform. Copyright (c) 2005 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * ---------------------------------------------------------------------------
 * ---------------------------------------------------------------------------
 */
package org.ow2.petals.component.framework.notification.notify;

import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.logging.Logger;

import javax.jbi.messaging.MessagingException;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactoryConfigurationError;
import javax.xml.xpath.XPathExpressionException;

import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.api.notification.NotificationBrokerController;
import org.ow2.petals.component.framework.api.notification.notify.NotifyController;

import com.ebmwebsourcing.wsstar.addressing.definition.api.EndpointReferenceType;
import com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.Subscribe;
import com.ebmwebsourcing.wsstar.notification.definition.utils.WSNotificationException;
import com.ebmwebsourcing.wsstar.notification.extension.utils.WSNotificationExtensionException;

/**
 * 
 * 
 * @author Frederic Gardes
 * 
 */
public class ComponentNotifyController implements NotifyController {
    /**
     * each EndpointReferenceType have a list of notifications generated by the
     * framework
     */
    private Map<EndpointReferenceType, List<AbstractProducerComponentNotify>> subscriptionMap = null;

    private NotificationBrokerController notificationBrokerController = null;

    private Logger logger = null;

    public ComponentNotifyController(
            final NotificationBrokerController notificationBrokerController, final Logger logger) {
        super();
        this.notificationBrokerController = notificationBrokerController;
        this.subscriptionMap = new ConcurrentHashMap<EndpointReferenceType, List<AbstractProducerComponentNotify>>();
        this.logger = logger;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.notification.notify.NotifyController
     * #add
     * (com.ebmwebsourcing.wsstar.addressing.definition.api.EndpointReferenceType
     * ,com.ebmwebsourcing.wsstar.notification.definition.basenotification.api.
     * Subscribe)
     */
    public void add(final EndpointReferenceType subscriptionReference, final Subscribe subscribe)
    throws WSNotificationException, WSNotificationExtensionException,
    XPathExpressionException, TransformerFactoryConfigurationError, TransformerException,
    ParserConfigurationException {
        boolean notify = false;
        List<AbstractProducerComponentNotify> componentNotifyList = null;
        // create the candidate component notify list
        final List<AbstractProducerComponentNotify> candidateAbstractProducerComponentNotifyList = this
        .createCandidateAbstractProducerComponentNotifyList(subscriptionReference,
                subscribe);
        // add the component notify's
        if (subscribe != null) {
            if (subscribe.getFilter() != null) {
                for (final AbstractProducerComponentNotify abstractProducerComponentNotify : candidateAbstractProducerComponentNotifyList) {
                    // check filter
                    // topic expression
                    notify = abstractProducerComponentNotify.getTopicExpressionFilter()
                    .isNotifiable(abstractProducerComponentNotify.getTopic());
                    if (notify) {
                        // create the component notify list if needed
                        if (componentNotifyList == null) {
                            componentNotifyList = new CopyOnWriteArrayList<AbstractProducerComponentNotify>();
                        }
                        componentNotifyList.add(abstractProducerComponentNotify);
                    }
                }
                if (componentNotifyList != null) {
                    this.subscriptionMap.put(subscriptionReference, componentNotifyList);
                }
            }
        }
    }

    /**
     * create the candidate component notify list
     * 
     * @param subscriptionReference
     * @param subscribe
     * @return a list of the candidates
     * @throws WSNotificationException
     * @throws ParserConfigurationException
     * @throws TransformerException
     * @throws TransformerFactoryConfigurationError
     * @throws XPathExpressionException
     * @throws WSNotificationExtensionException
     */
    private List<AbstractProducerComponentNotify> createCandidateAbstractProducerComponentNotifyList(
            final EndpointReferenceType subscriptionReference, final Subscribe subscribe)
            throws WSNotificationException, WSNotificationExtensionException,
            XPathExpressionException, TransformerFactoryConfigurationError, TransformerException,
            ParserConfigurationException {
        // TODO use a static way, to check on attributes instead of instance
        final List<AbstractProducerComponentNotify> candidateComponentNotifyList = new CopyOnWriteArrayList<AbstractProducerComponentNotify>();
        candidateComponentNotifyList.add(new InProducerComponentNotify(subscriptionReference,
                subscribe, this.notificationBrokerController, this.logger));
        candidateComponentNotifyList.add(new OutProducerComponentNotify(subscriptionReference,
                subscribe, this.notificationBrokerController, this.logger));
        candidateComponentNotifyList.add(new FaultProducerComponentNotify(subscriptionReference,
                subscribe, this.notificationBrokerController, this.logger));
        candidateComponentNotifyList.add(new StatusProducerComponentNotify(subscriptionReference,
                subscribe, this.notificationBrokerController, this.logger));
        return candidateComponentNotifyList;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.notification.notify.NotifyController
     * #remove
     * (com.ebmwebsourcing.wsstar.addressing.definition.api.EndpointReferenceType
     * )
     */
    public void remove(final EndpointReferenceType endpointReferenceType) {
        this.subscriptionMap.remove(endpointReferenceType);
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.component.framework.notification.notify.NotifyController
     * #process(java.lang.Class,
     * org.ow2.petals.component.framework.api.message.Exchange)
     */
    public void process(final Class<?> clazz, final Exchange exchange) {
        try {
            final Iterator<List<AbstractProducerComponentNotify>> abstractProducerComponentNotifyListIterator = this.subscriptionMap
            .values().iterator();
            while (abstractProducerComponentNotifyListIterator.hasNext()) {
                for (final AbstractProducerComponentNotify abstractProducerComponentNotify : abstractProducerComponentNotifyListIterator
                        .next()) {
                    if (clazz.isInstance(abstractProducerComponentNotify)) {
                        abstractProducerComponentNotify.process(exchange);
                    }
                }
            }
        } catch (final MessagingException e) {
            this.logger.warning("unable to check the notification list");
            this.logger.fine("detail on the problem during the check of the notification list:"
                    + e.getMessage());
        }
    }

}
