/**
 * PETALS - PETALS Services Platform. Copyright (c) 2008 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.component.framework.process;

import java.util.concurrent.ThreadPoolExecutor;
import java.util.logging.Level;
import java.util.logging.Logger;

import javax.jbi.messaging.DeliveryChannel;

import org.apache.commons.pool.impl.GenericObjectPool;
import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.listener.AbstractJBIListener;
import org.ow2.petals.component.framework.message.ExchangeImpl;

/**
 * The JBI Message Exchange Acceptor. Dispatch the message to a JBI worker which
 * will use the {@link AbstractJBIListener} component implementation.
 * 
 * @author chamerling - EBM WebSourcing
 * @since 3.0
 * 
 */
public class MessageExchangeAcceptor extends Thread {

    /**
     * The component's logger
     */
    private Logger logger;

    private DeliveryChannel deliveryChannel;

    private boolean running = false;

    private ThreadPoolExecutor jbiProcessorThreadPool;

    private GenericObjectPool jbiProcessorObjectPool;

    /**
     * 
     * @param component
     * @param runtimeConfiguration
     * @param jbiPool
     */
    public MessageExchangeAcceptor(AbstractComponent component,
            ThreadPoolExecutor jbiProcessorThreadPool, GenericObjectPool jbiProcessorObjectPool,
            Logger logger, ThreadGroup threadGroup, String threadName) {
        super(threadGroup, threadName);

        this.jbiProcessorThreadPool = jbiProcessorThreadPool;
        this.jbiProcessorObjectPool = jbiProcessorObjectPool;
        this.deliveryChannel = component.getChannel();
        this.logger = logger;
    }

    /*
     * (non-Javadoc)
     * 
     * @see java.lang.Thread#run()
     */
    @Override
    public void run() {
        this.logger.log(Level.FINE, "Starting MessageExchangeAcceptor Thread " + this.getName());
        this.running = true;

        while (this.running) {
            try {
                // receiving a message
                ExchangeImpl exchange = new ExchangeImpl(this.deliveryChannel.accept());

                this.logger.fine("Accepting a JBI message with Id : " + exchange.getExchangeId());

                MessageExchangeProcessor messageExchangeProcessor = (MessageExchangeProcessor) this.jbiProcessorObjectPool
                        .borrowObject();
                messageExchangeProcessor.setMessageExchange(exchange);
                this.jbiProcessorThreadPool.execute(messageExchangeProcessor);
            } catch (PEtALSCDKException e) {
                this.logger.log(Level.SEVERE, e.getMessage(), e);
            } catch (Exception e) {
                // search for an interrupted exception
                Throwable cause = e;
                while (cause != null && this.running) {
                    if (cause instanceof InterruptedException) {
                        this.running = false;
                        this.logger.fine("Thread '" + this.getName()
                                + "' interrupted, stop accepting JBI messages.");
                    }
                    cause = cause.getCause();
                }
                if (this.running) {
                    this.logger.info("Thread '" + this.getName()
                            + "' unexpectedly stopped !");
                }
            }
        }
    }
}
