/**
 * PETALS - PETALS Services Platform. Copyright (c) 2005 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * ---------------------------------------------------------------------------
 * ---------------------------------------------------------------------------
 */
package org.ow2.petals.component.framework.util;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.Random;

import org.ow2.petals.component.framework.api.util.IdUtil;

/**
 * This class provides a method to generate a unic id on a fast way
 * <p>
 * See {@link #getNewID()} for more information on the structure of the
 * generated UUID.
 * </p>
 * 
 * @author chamerling - eBM WebSourcing
 * @author cdeneux - Capgemini Sud
 * @author Frederic Gardes
 * 
 */
public class IdUtilImpl implements IdUtil {

    /**
     * Random value size.
     */
    private final static int RANDOM_VALUE_SIZE = 18;

    /**
     * The UUID namespace.
     */
    /**
     * 
     */
    private final static String UUID_NAMESPACE = "petals:uid:";

    /**
     * The unique identifier prefix.
     */
    private final String prefix;

    /**
     * The max size of the UUID string.
     */
    private final int UUID_SIZE = UUID_NAMESPACE.length() + RANDOM_VALUE_SIZE
    + String.valueOf(System.nanoTime()).length() + Long.toString(Long.MAX_VALUE).length();

    /**
     * The counter value used as identifier suffix.
     */
    private long value;

    /**
     * The mutex object used to synchronize the counter updates.
     */
    private final Object valueLock;

    /**
     * The constructor
     */
    public IdUtilImpl() {
        super();
        final Random myRand = new Random();

        final long rand = myRand.nextLong();
        String sid;
        try {
            sid = InetAddress.getLocalHost().toString();
        } catch (final UnknownHostException e) {
            sid = Thread.currentThread().getName();
        }
        final StringBuffer sb = new StringBuffer();
        sb.append(sid);
        sb.append(":");
        sb.append(Long.toString(rand));

        byte[] array;
        try {
            final MessageDigest md5 = MessageDigest.getInstance("MD5");
            md5.update(sb.toString().getBytes());
            array = md5.digest();
        } catch (final NoSuchAlgorithmException e) {
            // if the MD5 hasher is not reachable, just get the string directly
            array = sb.toString().getBytes();
        }

        final StringBuffer sb2 = new StringBuffer();
        for (int j = 0; j < array.length; ++j) {
            final int b = array[j] & 0xFF;
            sb2.append(Integer.toHexString(b));
        }
        int begin = myRand.nextInt();
        if (begin < 0) {
            begin = begin * -1;
        }
        begin = begin % 8;
        this.value = 0L;
        this.valueLock = new Object();
        this.prefix = UUID_NAMESPACE
        + sb2.toString().substring(begin, begin + RANDOM_VALUE_SIZE).toUpperCase();
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.component.framework.util.IdUtil#createId()
     */
    public String createId() {
        return this.createId(true);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.component.framework.api.util.IdUtil#createId(boolean)
     */
    public final String createId(final boolean prefixed) {
        final long tmp;
        synchronized (this.valueLock) {
            tmp = ++this.value;
            if (tmp >= Long.MAX_VALUE) {
                this.value = 0;
            }
        }
        final StringBuilder stringBuilder = new StringBuilder(this.UUID_SIZE);
        if (prefixed) {
            stringBuilder.append(this.prefix);
        }
        // We could use a StringBuilder because it is used only into the same
        // thread
        return stringBuilder.append(System.nanoTime()).append(
                tmp).toString();
    }
}
